#!/usr/bin/env python

import requests
from datetime import datetime

BASE_URL = "https://api.github.com"


def fetch_at(full_name: str):
    full_url = "{base_url}/repos/{full_name}".format(base_url=BASE_URL, full_name=full_name)
    return fetch(full_url)


def fetch(url: str):
    with requests.get(url) as response:
        json = response.json()
        if response.status_code is not 200:
            json['status_code'] = response.status_code
            timestamp = int(response.headers['X-RateLimit-Reset'])
            json['X-RateLimit-Reset'] = datetime.fromtimestamp(timestamp).isoformat()
        return response.status_code, json


def search_by(repo_name: str):
    query = "?q={repo_name}&order=desc".format(repo_name=repo_name)
    full_url = "{base_url}/search/repositories{query}".format(base_url=BASE_URL, query=query)
    return fetch(full_url)



#!/usr/bin/env python

import numbers


def convert_as_table_format(column_names, contents, is_headless=False, line_separator='\n', separator='|'):
    accumulation = []

    if not is_headless:
        table_header = convert_as_row_format(column_names)
        divider = prepare_divider(separator, len(column_names))
        accumulation.extend([table_header, divider])

    rows = [convert_as_row_format(row_items, separator) for row_items in contents]
    accumulation.extend(rows)
    return line_separator.join(accumulation)


def prepare_divider(separator, times):
    divider_tail = ' :----: {0}'.format(separator) * times
    return '{0}{1}'.format(separator, divider_tail)


def convert_as_row_format(items, separator='|'):
    margin_separator = ' {0} '.format(separator)
    body = margin_separator.join(beautify_text(item, separator) for item in items)
    return "{0} {1} {2}".format(separator, body, separator)


def beautify_text(text, separator):
    escaped_separator = '\{0}'.format(separator)
    return "{:,}".format(text) if isinstance(text, numbers.Number) else text.replace(separator, escaped_separator)



from unittest import TestCase

from girepo.arg_parser import create_argparser


class TestCreateArgparser(TestCase):
    def test_create_argparser_without_argument_error(self):
        test_data = [
            ({
                 "args": ["strict", "angular/angular", "facebook/react", "vuejs/vue"],
                 "sort_keys": ["star", "created_at"],
                 "expected": {
                     "names": ["angular/angular", "facebook/react", "vuejs/vue"],
                     "asc_key": None,
                     "desc_key": None,
                 }
             }, "normal arguments without sort option"),
            ({
                 "args": ["strict", "angular/   angular", "facebook   /react", "  vuejs  /  vue  "],
                 "sort_keys": ["star", "created_at"],
                 "expected": {
                     "names": ["angular/angular", "facebook/react", "vuejs/vue"],
                     "asc_key": None,
                     "desc_key": None,
                 }
             }, "arguments which have spaces without sort option"),
            ({
                 "args": ["strict", "angular/  angular", "facebook  /react", " vuejs / vue ", "--asc", "star"],
                 "sort_keys": ["star", "created_at"],
                 "expected": {
                     "names": ["angular/angular", "facebook/react", "vuejs/vue"],
                     "asc_key": "star",
                     "desc_key": None,
                 }
             }, "normal arguments with asc option"),
            ({
                 "args": ["strict", "angular/  angular", "facebook  /react", " vuejs / vue ", "--desc", "star"],
                 "sort_keys": ["star", "created_at"],
                 "expected": {
                     "names": ["angular/angular", "facebook/react", "vuejs/vue"],
                     "asc_key": None,
                     "desc_key": "star",
                 }
             }, "normal arguments with desc option"),

            # need to prepare data....
        ]

        for data, message in test_data:
            with self.subTest(message):
                actual = create_argparser(data['args'], data['sort_keys'])
                self.assertEqual(data['expected']['names'], actual.names)
                self.assertEqual(data['expected']['asc_key'], actual.asc_key)
                self.assertEqual(data['expected']['desc_key'], actual.desc_key)

    def test_create_argparser_with_argument_error(self):
        test_data = [
            ({
                "args": ["strict", "angular/  angular", "--sort"],
                "sort_keys": ["star", "created_at"],
                "expected": {
                    "full_names": ["angular/angular", "facebook/react", "vuejs/vue"],
                    "sort_key": None,
                    "is_asc": False,
                    "is_desc": False,
                }
            }, "There is no such option like sort"),
            ({
                "args": ["strict", "angular/  angular", "facebook  /react", " vuejs / vue ", "-a"],
                "sort_keys": ["star", "created_at"],
                "expected": {
                    "full_names": ["angular/angular", "facebook/react", "vuejs/vue"],
                    "sort_key": "star",
                    "is_asc": True,
                    "is_desc": False,
                }
            }, "the asc option requires a keyword in choices"),
            # need to prepare data....
        ]

        for data, message in test_data:
            with self.subTest(message):
                with self.assertRaises(SystemExit):
                    create_argparser(data['args'], data['sort_keys'])



from unittest import TestCase

from girepo.formatter import beautify_text


class TestFormatter(TestCase):

    def test_beautify_text(self):
        test_data = [
            ("Hello World", "|", "Hello World"),
            ("Hello | World", "|", "Hello \| World"),
            ("Hello $ World", "$", "Hello \$ World"),
            ("Hello | World", "$", "Hello | World")
        ]

        for target, separator, expected in test_data:
            with self.subTest():
                actual = beautify_text(target, separator)
                self.assertEqual(expected, actual)



#!/usr/bin/env python

import logging
import sys

from girepo.arg_parser import create_argparser, is_strict_command, is_rough_command
from girepo.formatter import convert_as_table_format
from girepo.extractor import extract, default_mapper
from girepo.fetch import fetch_at, search_by
from girepo.sort import sort_by


def retrieve_data_directly(repo_full_names, mapper):
    contents = []
    for repo_full_name in repo_full_names:
        status_code, json = fetch_at(repo_full_name)
        add_retrieved_data(contents, json, mapper, status_code)

    return contents


def add_retrieved_data(contents, repo_json, mapper, status_code):
    if status_code == 200:
        content = extract(repo_json, mapper)
        contents.append(content)
    else:
        logging.error(repo_json)


def retrieve_data_heuristic(repo_names, mapper):
    contents = []
    for repo_name in repo_names:
        status_code, json = search_by(repo_name)
        if json.get("items"):
            add_retrieved_data(contents, json['items'][0], mapper, status_code)

    return contents


def main():
    column_names = list(default_mapper.keys())
    parser = create_argparser(sys.argv[1:], column_names)

    if is_strict_command(parser):
        contents = retrieve_data_directly(parser.names, default_mapper)
    elif is_rough_command(parser):
        contents = retrieve_data_heuristic(parser.names, default_mapper)
    else:
        sys.exit("missing a sub-command. see the usage `girepo --help`")

    contents = sort_by(column_names, contents, parser)

    table_format = convert_as_table_format(column_names, contents, parser.headless)

    print(table_format)


if __name__ == '__main__':
    main()



from unittest import TestCase

from girepo.extractor import calculate_star_per


class TestExtractor(TestCase):

    def test_calculate_star_per(self):
        created_at, updated_at, star_count, day_duration, round_n = \
            ["2014-09-11T17:12:01Z", "2014-09-18T16:12:01Z", 14, 1, 2]
        actual = calculate_star_per(created_at, updated_at, star_count, day_duration, round_n)
        self.assertEqual(2.00, actual)

    def test_calculate_star_per_case_zero_division(self):
        created_at, updated_at, star_count, day_duration, round_n = \
            ["2014-09-18T16:12:01Z", "2014-09-18T16:12:01Z", 0, 1, 2]
        actual = calculate_star_per(created_at, updated_at, star_count, day_duration, round_n)
        self.assertEqual(0.00, actual)

    def test_calculate_star_per_case_same_date(self):
        created_at, updated_at, star_count, day_duration, round_n = \
            ["2014-09-18T16:12:01Z", "2014-09-18T16:12:01Z", 1, 1, 2]
        actual = calculate_star_per(created_at, updated_at, star_count, day_duration, round_n)
        self.assertEqual(1.00, actual)



#!/usr/bin/env python


def sort_by(keys, contents, parser):
    """
    the default sort order is desc.
    :param keys: should be a list which has the index method.
    :param contents:
    :param parser:
    :return: new list
    """

    if parser.asc_key:
        index = keys.index(parser.asc_key)
        return sorted(contents, key=lambda x: x[index])
    elif parser.desc_key:
        index = keys.index(parser.desc_key)
        return sorted(contents, key=lambda x: x[index], reverse=True)

    return contents.copy()



#!/usr/bin/env python
from datetime import datetime
from collections import OrderedDict


def calculate_star_per(created_at, updated_at, star_count, day_duration=1, round_n=2):
    """

    :param created_at: e.g. 2014-09-18T16:12:01Z
    :param updated_at: e.g. 2014-09-19T16:12:01Z
    :param star_count:
    :param day_duration: if 1, it means per day. if 7, it means per week.
    :param round_n:
    :return:  day_duration * start / (days + 1)
    """
    delta = datetime.strptime(updated_at, "%Y-%m-%dT%H:%M:%SZ") - datetime.strptime(created_at, "%Y-%m-%dT%H:%M:%SZ")
    offset_for_zero_division = 1
    return day_duration * round(star_count / (delta.days + offset_for_zero_division), round_n)


def to_date_format(date_str):
    return date_str.split('T')[0]


def to_link_format(url):
    return "[link]({0})".format(url)


def to_fullname_format(owner, name):
    return "{0}/{1}".format(owner, name)


def to_license_name(license_obj):
    return license_obj['name'] if license_obj else "Unknown, see Homepage"


default_mapper = OrderedDict([
    ('fullname', lambda x: to_fullname_format(
        (x['owner']['login']), x['name'])
     ),
    ('star', lambda x: x['stargazers_count']),
    ('star/day', lambda x: calculate_star_per(
        x['created_at'], x['updated_at'], x['stargazers_count'], day_duration=1)
     ),
    ('created_at', lambda x: to_date_format(x['created_at'])),
    ('updated_at', lambda x: to_date_format(x['updated_at'])),
    ('license', lambda x: to_license_name(x['license'])),
    ('language', lambda x: x['language']),
    ('description', lambda x: x['description']),
    ('url', lambda x: to_link_format(x['html_url'])),
])


def extract(data, mapper):
    return [func(data) for _, func in mapper.items()]



#!/usr/bin/env python

import re
import argparse

from enum import Enum


class SubParser(Enum):
    STRICT = "strict"
    ST = "st"
    ROUGH = "rough"
    RO = "ro"


def is_strict_command(parser):
    return parser.command_name == SubParser.STRICT.value or parser.command_name == SubParser.ST.value


def is_rough_command(parser):
    return parser.command_name == SubParser.ROUGH.value or parser.command_name == SubParser.RO.value


def full_name_type(value, pattern=re.compile(r"^(( )*[\w.\-]+( )*/( )*[\w.\-]+( )*)$")):
    """
    allow the value contains white-space at around words

    https://regex101.com/r/ks59iR/4

    :param value:
    :param pattern:
    :return:
    """
    if not pattern.match(value):
        raise argparse.ArgumentTypeError("the format of name should be like owner/repo")
    return value.replace(' ', '')


def create_argparser(args, sort_keys):
    parser = argparse.ArgumentParser(
        description="""
        Print Git Repositories information with the markdown format for your documents.
        
        This uses github api without authentication. 
        So you could reach the rate-limitation if you use this 60 requests per hour.
        https://developer.github.com/v3/#rate-limiting
        
        Could you take a cup of coffee \U00002615 until recovering if you exceed. 
        """,
        epilog="God bless you.",
        fromfile_prefix_chars="@"
    )

    subparsers_action = parser.add_subparsers(
        title='subcommands',
        description="you can choose a search way from sub-commands.\
         rough sub-command does not require owner info but it might return wrong info.",
        dest='command_name',
    )

    # rough
    subparser_heuristic = add_heuristic_subparser(subparsers_action)
    add_common_options(subparser_heuristic, sort_keys)

    # strict
    subparser_strict = add_strict_subparser(subparsers_action)
    add_common_options(subparser_strict, sort_keys)

    return parser.parse_args(args)


def add_strict_subparser(subparsers_action):
    subparser_strict = subparsers_action.add_parser(SubParser.STRICT.value,
                                                    description="This is a strict search.\
                                               This requires owner and repository name as \"onwer/repository\".",
                                                    aliases=[SubParser.ST.value],
                                                    help='strict search. see `girepo st --help`')
    subparser_strict.add_argument('names', type=full_name_type, nargs='+',
                                  help='the target repositories written like owner/repository')
    return subparser_strict


def add_heuristic_subparser(subparsers_action):
    subparser_heuristic = subparsers_action.add_parser(SubParser.ROUGH.value,
                                                       description="This is a heuristic search.\
                                                 This is possible to return the wrong repository info.",
                                                       aliases=[SubParser.RO.value],
                                                       help='heuristic search. see `girepo ro --help`')
    subparser_heuristic.add_argument('names', nargs='+',
                                     help='the target repositories')
    return subparser_heuristic


def add_common_options(parser, sort_keys):
    parser.add_argument('--headless', dest="headless", action='store_true',
                        help='not to describe table headers')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-a', '--asc', dest="asc_key", choices=sort_keys,
                       help='sort by asc with the field')
    group.add_argument('-d', '--desc', dest="desc_key", choices=sort_keys,
                       help='sort by desc with the field')
